# -*- coding: utf-8 -*-
from __future__ import unicode_literals, division, print_function, absolute_import
import inspect
import sys
import datetime

from datatypes import (
    Pool,
    EnglishWord,
    NamingConvention,
    classproperty
)

# first party
from .query import Query, Iterator
from . import utils
from .interface import get_interface
from .config import (
    Schema,
    Field,
    AutoDatetime,
    AutoIncrement,
    Index
)
from .compat import *


class OrmPool(Pool):
    """
    Create a pool of Orm instances, which is just a dict of primary_key -> Orm instance
    mappings

    Let's say you are iterating through millions of rows of Foo, and for each Foo
    instance you need to get the Bar instance from the Foo.bar_id field, and lots of
    Foos have the same bar_id, but you only want to pull the Bar instance from
    the db once, this allows you to easily do that

    :Example:
        bar_pool = Bar.pool(500) # keep the pool contained to the last 500 Bar instances
        for f in Foo.query.all():
            b = bar_pool[f.bar_id]
            print "Foo {} loves Bar {}".format(f.pk, b.pk)
    """
    def __init__(self, orm_class, maxsize=0):
        super(OrmPool, self).__init__(maxsize=maxsize)
        self.orm_class = orm_class

    def __missing__(self, pk):
        o = self.orm_class.query.eq_pk(pk).one()
        self[pk] = o
        return o


class Orm(object):
    """
    this is the parent class of any Orm child class you want to create that can access the db

    :example:
        from prom import Orm, Field, Index

        # create a simple class using standard fields
        class Foo(Orm):
            table_name = "<TABLE NAME>"

            bar = Field(int, True, max_size=512, default=0, unique=True)
            che = Field(str, True)

            index_barche = Index('bar', 'che')
    """
    connection_name = ""
    """the name of the connection to use to retrieve the interface"""

    query_class = Query
    """the class this Orm will use to create Query instances to query the db"""

    iterator_class = Iterator
    """the class this Orm will use for iterating through results returned from db"""

    orm_classes = {}
    """This will hold all other orm classes that have been loaded into memory
    the class path is the key and the class object is the value"""

    _id = AutoIncrement(aliases=["id"])
    _created = AutoDatetime(created=True, updated=False, aliases=["created"])
    _updated = AutoDatetime(created=False, updated=True, aliases=["updated"])

    @classproperty
    def table_name(cls):
        return NamingConvention("{}_{}".format(
            cls.__module__.replace(".", "_"),
            cls.__name__
        )).snakecase()

    @classproperty
    def model_name(cls):
        """Returns the name for this orm

        :Example:
            class Collection(Orm): pass

            print(Collection.model_name) # collection
            print(Collection.models_name) # collections

        :param plural: bool, True if you would like the plural model name
        :returns: str, the model name in either singular or plural form
        """
        return NamingConvention(cls.__name__).snakecase()

    @classproperty
    def models_name(cls):
        """Returns the plural name for this orm

        :see: .model_name
        :returns: str, the model name in plural form
        """
        return EnglishWord(cls.model_name).plural()

    @classproperty
    def schema(cls):
        """the Schema() instance that this class will derive all its db info from"""
        return Schema.get_instance(cls)

    @classproperty
    def interface(cls):
        """
        return an Interface instance that can be used to access the db

        return -- Interface() -- the interface instance this Orm will use
        """
        return get_interface(cls.connection_name)

    @classproperty
    def query(cls):
        """return a new Query instance ready to make a db call using the child class

        :example:
            # fluid interface
            results = Orm.query.is_foo('value').desc_bar().get()

        :returns: Query, every time this is called a new query instance is created using cls.query_class
        """
        query_class = cls.query_class
        return query_class(orm_class=cls)

    @property
    def field_names(self):
        """Return all the field names"""
        return [k for k in self.schema.fields]

    @property
    def fields(self):
        """
        return all the fields and their raw values for this Orm instance. This
        property returns a dict with the field names and their current values

        if you want to control the values for outputting to an api, use .jsonable()
        """
        return {k:getattr(self, k, None) for k in self.schema.fields}

    @property
    def modified_field_names(self):
        """Return all the field names that are currently modified"""
        modified_field_names = set()
        for field_name, field in self.schema.fields.items():
            if field.modified(self, getattr(self, field_name)):
                modified_field_names.add(field_name)
        return modified_field_names

    @property
    def modified_fields(self):
        """Return a dict of field_names/field_values for all the currently modified fields"""
        return {k:getattr(self, k) for k in self.modified_field_names}

    @classmethod
    def pool(cls, maxsize=0):
        """
        return a new OrmPool instance

        return -- OrmPool -- the orm pool instance will be tied to this Orm
        """
        return OrmPool(orm_class=cls, maxsize=maxsize)

    @classmethod
    def transaction(cls, **kwargs):
        """Create a transaction for this Orm

        :Example:
            with FooOrm.transaction() as conn:
                o = FooOrm(foo=1)
                o.save(connection=conn)

        :param **kwargs: passed through to the Interface.transaction context manager
            * prefix: str, the name of the transaction you want to use
            * nest: bool, True if you want nested transactions to be created, False
                to ignore nested transactions
        :returns: Connection instance
        """
        kwargs.setdefault("nest", False)
        kwargs.setdefault("prefix", f"{cls.__name__}_{cls.connection_name}_tx")
        return cls.interface.transaction(**kwargs)

    @classmethod
    def create(cls, *args, **kwargs):
        """
        create an instance of cls with the passed in fields and set it into the db

        this method takes *args, **kwargs because a child class can override .__init__
        and it's nice to not have to modify this method also

        :param *args: passed directly to .__init__
        :param **kwargs: passed directly to .__init__
        :returns: Orm instance that has been saved into the db
        """
        # NOTE -- you cannot use hydrate/populate here because populate alters modified fields
        connection = kwargs.pop("connection", None)

        instance = cls(*args, **kwargs)

        if connection:
            instance.save(connection=connection)

        else:
            instance.save()

        return instance

    @classmethod
    def hydrate(cls, fields=None, **fields_kwargs):
        """return a populated instance with the present fields

        NOTE -- you probably shouldn't override this method since the Query methods
        rely on this method signature to create each instance of the results

        :param fields: dict, the fields to populate in this instance
        :param **fields_kwargs: dict, the fields in key=val form to populate in this instance
        :returns: an instance of this class with populated fields
        """
        instance = cls()
        fields = cls.make_dict(fields, fields_kwargs)
        instance.from_interface(fields)
        instance._interface_hydrate = True
        return instance

    @classmethod
    def make_dict(cls, fields, fields_kwargs, schema=None):
        """Lots of methods take a dict and key=val for fields, this combines fields
        and fields_kwargs into one master dict, turns out we want to do this more
        than I would've thought to keep api compatibility with prom proper

        :param fields: dict, the fields in a dict
        :param fields_kwargs: dict, if you would like to pass the fields as key=val
            this picks those up and combines them with fields
        :schema: Schema, if passed in then this will normalize field names
        :returns: dict, the combined fields
        """
        fields = utils.make_dict(fields, fields_kwargs)

        if schema:
            # since schema is passed in resolve any aliases
            for field_name in list(fields.keys()):
                if fn := schema.field_name(field_name, ""):
                    fields[fn] = fields.pop(field_name)

        return fields

    def __init__(self, fields=None, **fields_kwargs):
        """Create an Orm object

        While you can override this method to customize the signature, you might
        also need to override .hydrate (but don't change .hydrate's signature)
        since .hydrate creates an instance using no arguments

        :param fields: dict, the fields in a dict
        :param **fields_kwargs: dict, if you would like to pass the fields as key=val
        """
        self._interface_pk = None
        self._interface_hydrate = False

        schema = self.schema
        fields = self.make_dict(fields, fields_kwargs, schema=schema)

        # set defaults
        for field_name, field in schema.fields.items():
            fields[field_name] = field.fdefault(self, fields.get(field_name, None))

        self.modify(fields)

    def __init_subclass__(cls):
        """When this class is loaded into memory it will be saved into cls.orm_classes,
        this way every orm class knows about all the others

        https://peps.python.org/pep-0487/
        """
        super().__init_subclass__()
        cls.orm_classes[f"{cls.__module__}:{cls.__name__}"] = cls

    def fk(self, orm_class):
        """find the field value in self that is the primary key of the passed in orm_class

        :example:
            class Foo(Orm):
                pass

            class Bar(Orm):
                foo_id = Field(Foo)

            b = Bar(foo_id=1)
            print(b.fk(Foo)) # 1

        :param orm_class: Orm, the fields in self will be checked until the field that
            references Orm is found, then the value of that field will be returned
        :returns: the self field value that is a foreign key references to orm_class
        """
        for field_name, field in self.schema.ref_fields.items():
            if field.schema is orm_class.schema:
                return getattr(self, field_name)

        raise ValueError("Did not find a foreign key reference for {} in {}".format(
            orm_class.__name__,
            self.__class__.__name__,
        ))

    def ref(self, orm_classpath):
        """see Query.ref() for an explanation of what this method does

        :param orm_classpath: string|type, a full python class path (eg, foo.bar.Che) or
            an actual model.Orm python class
        :returns: Orm
        """
        return self.query.ref(orm_classpath).orm_class

    def ref_class(self, orm_classpath):
        return self.ref(orm_classpath)

    def is_hydrated(self):
        """return True if this orm was populated from the interface/db"""
        return self._interface_hydrate

    def is_update(self):
        """Return True if .save() will perform an interface update"""
        return self._interface_pk is not None

    def is_insert(self):
        """Return True if .save() will perform an interface insert"""
        return not self.is_update()

    def from_interface(self, fields):
        """this runs all the fields through their iget methods to mimic them
        freshly coming out of the db, then resets modified

        :param fields: dict, only the fields you want to populate
        """
        schema = self.schema
        for field_name, v in fields.items():
            if field_name in schema.fields:
                fields[field_name] = schema.fields[field_name].iget(self, v)

        self.modify(fields)

        # this marks that this was repopulated from the interface (database)
        self._interface_pk = self.pk

    def to_interface(self):
        """Get all the fields that need to be saved

        :returns: dict[str, Any], key is field_name and val is the field value
            to be saved
        """
        fields = {}
        schema = self.schema
        for k, field in schema.fields.items():
            v = field.iset(self, getattr(self, k))

            is_modified = field.modified(self, v)
            if is_modified:
                if not field.options.get("empty", True) and not v:
                    raise ValueError(f"{self.__class__.__name__}.{k} cannot be empty")

                fields[k] = v

            if v is None and field.is_required():
                if field.is_pk():
                    if is_modified:
                        raise KeyError("Primary key has been removed and is required")

                else:
                    if self.is_insert() or is_modified:
                        raise KeyError(f"Missing required field {self.__class__.__name__}.{k}")

        return fields

    def insert(self, **kwargs):
        """persist the field values of this orm"""
        ret = True

        schema = self.schema
        fields = self.to_interface()

        q = self.query
        q.set(fields)
        pk = q.insert(**kwargs)
        if pk:
            fields = q.fields_set.fields
            pk_name = schema.pk_name
            if pk_name:
                fields[pk_name] = pk
                self.from_interface(fields)

        else:
            ret = False

        return ret

    def update(self, **kwargs):
        """re-persist the updated field values of this orm that has a primary key"""
        ret = True
        fields = self.to_interface()

        q = self.query
        q.set(fields)

        pk = self._interface_pk
        if pk:
            q.eq_field(self.schema.pk.name, pk)

        else:
            raise ValueError("Cannot update an unhydrated orm instance")

        if q.update(**kwargs):
            fields = q.fields_set.fields
            self.from_interface(fields)

        else:
            ret = False

        return ret

    def upsert(self, **kwargs):
        """Perform an UPSERT query where we insert the fields if they don't already
        exist on the db or we UPDATE if they do

        We only want to upsert on specific occasions where we know we've set the
        conflict values and will be sending them to the db. UPSERT queries need
        to have a unique index on the table they can use for the conflict fields.

        This method will go through the indexes and try and find a unique index
        that has all fields that are being sent to the interface and it will use
        those fields as the conflict fields, it will raise a ValueError if it can't
        find a valid set of conflict fields

        :param **kwargs: passed through to the interface
        """
        ret = True

        if pk := self._interface_pk:
            ret = self.update()

        else:
            schema = self.schema
            fields = self.to_interface()

            conflict_fields = self.conflict_fields(fields)
            if not conflict_fields:
                raise ValueError(
                    "{}.upsert() failed to find conflict field names from: {}".format(
                        self.__class__.__name__,
                        list(fields.keys())
                    )
                )

            q = self.query
            q.set(fields)
            pk = q.upsert([t[0] for t in conflict_fields], **kwargs)
            if pk:
                fields = q.fields_set.fields
                pk_name = schema.pk_name
                if pk_name:
                    fields[pk_name] = pk
                self.from_interface(fields)

            else:
                ret = False

        return ret

    def save(self, **kwargs):
        """
        persist the fields in this object into the db, this will update if _id is set, otherwise
        it will insert

        see also -- .insert(), .update()
        """
        ret = False

        pk = self._interface_pk
        if pk:
            ret = self.update(**kwargs)
        else:
            ret = self.insert(**kwargs)

        return ret

    def delete(self, **kwargs):
        """delete the object from the db if pk is set"""
        ret = False
        q = self.query
        pk = self._interface_pk
        if pk:
            pk_name = self.schema.pk.name
            self.query.eq_field(pk_name, pk).delete(**kwargs)

            for field_name, field in self.schema.fields.items():
                setattr(self, field_name, field.idel(self, getattr(self, field_name)))

            self._interface_pk = None
            self._interface_hydrate = False

            ret = True

        return ret

    def conflict_fields(self, fields):
        """Internal method. This will find fields that can be used for .upsert/.load

        :param fields: dict, the fields to check for values that would satisfy
            unique indexes or a primary key
        :returns: list<tuple>, a list of (field_name, field_value) tuples
        """
        conflict_fields = []

        schema = self.schema

        # we'll start with checking the primary key
        for field_name in schema.pk_names:
            if field_name in fields:
                conflict_fields.append((field_name, fields[field_name]))

            else:
                conflict_fields = []
                break

        if not conflict_fields:
            # no luck with the primary key, so let's check unique indexes
            for index in schema.indexes.values():
                if index.unique:
                    for field_name in index.field_names:
                        if field_name in fields:
                            conflict_fields.append((field_name, fields[field_name]))

                        else:
                            conflict_fields = []
                            break

                    if conflict_fields:
                        break

        return conflict_fields

    def load(self):
        """Given a partially populated orm try and load any missing fields from
        the db

        :returns: bool, True if it loaded from the db, False otherwise
        """
        fields = self.modified_fields
        conflict_fields = self.conflict_fields(fields)
        if not conflict_fields:
            raise ValueError(f"Load failed to find suitable fields to query on")

        q = self.query
        for field_name, field_val in conflict_fields:
            q.eq_field(field_name, field_val)

        field_names = []
        for field_name in self.schema.fields.keys():
            if field_name not in fields:
                field_names.append(field_name)

        q.select(*field_names)
        field_values = q.one()
        if field_values:
            ret = True
            fields = dict(zip(field_names, field_values))
            self.from_interface(fields)

            # can't decide if I should actually set this or not
            self._interface_hydrate = True

        else:
            ret = False

        return ret

    def requery(self):
        """Fetch this orm from the db again (ie, re-query the row from the db and
        return a new Orm instance with the columns from that row)"""
        fields = {k:v for k, v in self.fields.items() if v is not None}

        conflict_fields = self.conflict_fields(fields)
        if not conflict_fields:
            raise ValueError("Unable to refetch orm")

        q = self.query
        for field_name, field_val in conflict_fields:
            q.eq_field(field_name, field_val)

        return q.one()

    def is_modified(self, field_name=""):
        """true if a field, or any field, has been changed from its original value, false otherwise

        :param field_name: string, the name of the field you want to check for modification
        :returns: bool
        """
        if field_name:
            ret = field_name in self.modified_field_names
        else:
            ret = len(self.modified_field_names) > 0
        return ret

    def modify(self, fields=None, **fields_kwargs):
        """update the fields of this instance with the values in dict fields

        this should rarely be messed with, if you would like to manipulate the
        fields you should override modify_fields()

        :param fields: dict, the fields in a dict
        :param **fields_kwargs: dict, if you would like to pass the fields as key=val
            this picks those up and combines them with fields
        """
        fields = self.make_dict(fields, fields_kwargs)
        fields = self.modify_fields(fields)
        schema = self.schema

        for field_name, field_val in fields.items():
            if fn := schema.field_name(field_name, None):
                setattr(self, fn, field_val)

    def modify_fields(self, fields):
        """In child classes you should override this method to do any default 
        customizations on the fields, so if you want to set defaults or anything
        you should do that here

        :param fields: dict, the fields you might want to be modified
        :returns: dict, the fields you want to actually be modified
        """
        return fields

    def __getattr__(self, k):
        ret = None
        try:
            field_name = self.schema.field_name(k)

        except AttributeError:
            for ref_field_name, ref_field in self.schema.ref_fields.items():
                ref_class = ref_field.ref
                if k == ref_class.model_name:
                    ref_field_value = getattr(self, ref_field_name, None)
                    ret = None
                    if ref_field_value:
                        ret = ref_class.query.eq_pk(ref_field_value).one()

                    return ret

            # Go through all the orm_classes looking for a model_name or models_name
            # match and query that model using that model's FK field name that
            # matches self.pk
            for orm_class in self.orm_classes.values():
                if k == orm_class.models_name or k == orm_class.model_name:
                    for ref_field_name, ref_field in orm_class.schema.ref_fields.items():
                        ref_class = ref_field.ref
                        if ref_class and isinstance(self, ref_class):
                            query = orm_class.query.eq_field(ref_field_name, self.pk)
                            return query.all() if k == orm_class.models_name else query.one()

            raise

        else:
            ret = getattr(self, field_name)

        return ret

    def __setattr__(self, k, v):
        try:
            field_name = self.schema.field_name(k)

        except AttributeError:
            for ref_field_name, ref_field in self.schema.ref_fields.items():
                ref_class = ref_field.ref
                if k == ref_class.model_name:
                    k = ref_field_name
                    v = v.pk
                    break

            field_name = k
        return super().__setattr__(field_name, v)

    def __delattr__(self, k):
        try:
            field_name = self.schema.field_name(k)
        except AttributeError:
            field_name = k
        return super(Orm, self).__delattr__(field_name)

    def __int__(self):
        return int(self.pk)

    def __long__(self):
        return long(self.pk)

    def __str__(self):
        #return self.__unicode__() if is_py2 else self.__bytes__()
        return str(self.pk)

    def __unicode__(self):
        return unicode(self.pk)

    def __bytes__(self):
        return bytes(self.pk)

    def jsonable(self, *args, **options):
        """
        return a public version of this instance that can be jsonified

        Note that this does not return _id, _created, _updated, the reason why is
        because lots of times you have a different name for _id (like if it is a 
        user object, then you might want to call it user_id instead of _id) and I
        didn't want to make assumptions

        note 2, I'm not crazy about the name, but I didn't like to_dict() and pretty
        much any time I need to convert the object to a dict is for json, I kind of
        like dictify() though, but I've already used this method in so many places
        """
        d = {}
        for field_name, field in self.schema.fields.items():
            field_name, field_val = field.jsonable(
                self,
                field_name,
                getattr(self, field_name, None)
            )
            if field_val is not None:
                d[field_name] = field_val

        return d

    @classmethod
    def install(cls):
        """install the Orm's table using the Orm's schema"""
        return cls.interface.set_table(cls.schema)




# -*- coding: utf-8 -*-
"""
Bindings for SQLite

https://docs.python.org/2/library/sqlite3.html

Notes, certain SQLite versions might have a problem with long integers
http://jakegoulding.com/blog/2011/02/06/sqlite-64-bit-integers/

Looking at the docs, it says it will set an integer value to 1-4, 6, or 8 bytes
depending on the size, but I couldn't get it to accept anything over the 32-bit signed
integer value of around 2billion

savepoints and transactions are similar to Postgres
https://www.sqlite.org/lang_savepoint.html
http://sqlite.org/lang_transaction.html

alter table is similar to Postgres
https://www.sqlite.org/lang_altertable.html

other links that were helpful
http://www.numericalexpert.com/blog/sqlite_blob_time/
"""
from __future__ import unicode_literals, division, print_function, absolute_import
import os
import decimal
import datetime
from distutils import dir_util
import re
import sqlite3
import json
import uuid

from datatypes import Datetime

# first party
from ..exception import (
    InterfaceError,
    UniqueError,
    TableError,
    FieldError,
    UniqueError,
    CloseError,
    PlaceholderError,
)

from ..compat import *
from .sql import SQLInterface, SQLConnection


class SQLiteConnection(SQLConnection, sqlite3.Connection):
    """
    Thin wrapper around the default connection to make sure it has a similar interface
    to Postgres' connection instance so the common code can all be the same in the
    parent class

    https://docs.python.org/3.11/library/sqlite3.html#sqlite3.Connection
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.closed = 0

    def close(self, *args, **kwargs):
        r = super().close(*args, **kwargs)
        self.closed = 1
        return r


class BooleanType(object):
    FIELD_TYPE = 'BOOL'

    @classmethod
    def adapt(cls, val):
        """From python you get False and True, convert those to 1/0"""
        return 1 if val else 0

    @classmethod
    def convert(cls, val):
        """from the db you get values like b'0' and b'1', convert those to True/False"""
        return bool(int(val))


class StringType(object):
    """this just makes sure 8-bit bytestrings get converted ok"""
    FIELD_TYPE = 'TEXT'

    @classmethod
    def adapt(cls, val):
        return String(val)


class DatetimeType(StringType):
    """External sqlite3 databases can store the TIMESTAMP type as unix timestamps,
    this caused parsing problems when pulling the values out of the db because the
    default adapter expected TIMESTAMP to be in the form of YYYY-MM-DD HH:MM:SS.SSSSSS
    and so it would fail to convert the DDDDDD.DDD values, this handles that conversion

    https://www.sqlite.org/lang_datefunc.html
    the "unixepoch" modifier only works for dates between 0000-01-01 00:00:00 and
    5352-11-01 10:52:47 (unix times of -62167219200 through 106751991167)

    uses the name TIMESTAMP over DATETIME to be consistent with Postgres
    """
    FIELD_TYPE = 'TIMESTAMP'

    @classmethod
    def adapt(cls, val):
        return Datetime(val).isoformat()

    @classmethod
    def convert(cls, val):
        return Datetime(super().adapt(val)).datetime()
        #return Datetime(super().adapt(val)).datetime().replace(tzinfo=datetime.timezone.utc)


class NumericType(object):
    """numbers bigger than 64bit integers can be stored as this

    This is named to be as consistent with Postgres's NUMERIC(<precision>, 0) type

    This has TEXT in the name so it is treated as text according to SQLite's
    order of affinity rule 2:

        If the declared type of the column contains any of the strings "CHAR",
        "CLOB", or "TEXT" then that column has TEXT affinity.

        https://www.sqlite.org/datatype3.html
    """
    FIELD_TYPE = 'NUMERICTEXT'

    @classmethod
    def adapt(cls, val):
        if val < 9223372036854775807:
            return val
        else:
            v = str(val)
            return v

    @classmethod
    def convert(cls, val):
        """This should only be called when the column type is actually FIELD_TYPE"""
        return int(val)


class DecimalType(object):
    FIELD_TYPE = 'DECIMALTEXT'

    @staticmethod
    def adapt(val):
        return str(val)

    @staticmethod
    def convert(val):
        ret = decimal.Decimal(val)
        return ret


class DictType(object):
    """Converts a dict to json text and back again

    Uses JSONBTEXT to be as close to Postgres while still triggering SQLite's affinity rule 2
    """
    FIELD_TYPE = 'JSONBTEXT'

    @classmethod
    def adapt(cls, val):
        return json.dumps(val)

    @classmethod
    def convert(cls, val):
        """This should only be called when the column type is actually FIELD_TYPE"""
        return json.loads(val)


class SQLite(SQLInterface):
    """
    https://docs.python.org/3/library/sqlite3.html
    """
    LIMIT_NONE = -1

    _connection = None

    @classmethod
    def configure(cls, config):
        dsn = getattr(config, 'dsn', '')
        if dsn:
            host = config.host
            db = config.database
            if not host:
                path = db

            elif not db:
                path = host

            else:
                path = os.sep.join([host, db])

        else:
            path = config.database

        if not path:
            raise ValueError("no sqlite db path found in config")

        config.path = path
        return config

    def get_paramstyle(self):
        """
        https://docs.python.org/3/library/sqlite3.html#sqlite3.paramstyle
        """
        return sqlite3.paramstyle

    def _connect(self, config):
        """
        https://docs.python.org/3.11/library/sqlite3.html#sqlite3.connect
        """
        path = config.path

        # https://docs.python.org/2/library/sqlite3.html#default-adapters-and-converters
        options = {
            'isolation_level': None,
            #'isolation_level': "IMMEDIATE",
            #'isolation_level': "EXCLUSIVE",
            'detect_types': sqlite3.PARSE_DECLTYPES|sqlite3.PARSE_COLNAMES,
            'factory': SQLiteConnection,
            'check_same_thread': True, # https://stackoverflow.com/a/2578401/5006
            #'timeout': 100,
        }
        option_keys = list(options.keys()) + ['timeout', 'cached_statements']
        for k in option_keys:
            if k in config.options:
                options[k] = config.options[k]

        try:
            self._connection = sqlite3.connect(path, **options)
            self.log_debug(f"Connected to connection 0x{id(self._connection):02x}")

        except sqlite3.DatabaseError as e:
            path_d = os.path.dirname(path)
            if os.path.isdir(path_d):
                raise

            else:
                # let's try and make the directory path and connect again
                dir_util.mkpath(path_d)
                self._connection = sqlite3.connect(path, **options)

        # https://docs.python.org/2/library/sqlite3.html#row-objects
        self._connection.row_factory = sqlite3.Row
        # https://docs.python.org/2/library/sqlite3.html#sqlite3.Connection.text_factory
        self._connection.text_factory = StringType.adapt

        # for some reason this is needed in python 3.6 in order for saved bytes
        # to be ran through the converter, not sure why
        sqlite3.register_converter(StringType.FIELD_TYPE, StringType.adapt)

        sqlite3.register_adapter(bool, BooleanType.adapt)
        sqlite3.register_converter(BooleanType.FIELD_TYPE, BooleanType.convert)

        # sadly, it doesn't look like these work for child classes so each class
        # has to be adapted even if its parent is already registered
        sqlite3.register_adapter(datetime.datetime, DatetimeType.adapt)
        sqlite3.register_adapter(Datetime, DatetimeType.adapt)
        sqlite3.register_converter(DatetimeType.FIELD_TYPE, DatetimeType.convert)

        sqlite3.register_adapter(int, NumericType.adapt)
        sqlite3.register_converter(NumericType.FIELD_TYPE, NumericType.convert)

        sqlite3.register_adapter(decimal.Decimal, DecimalType.adapt)
        sqlite3.register_converter(DecimalType.FIELD_TYPE, DecimalType.convert)

        sqlite3.register_adapter(dict, DictType.adapt)
        sqlite3.register_converter(DictType.FIELD_TYPE, DictType.convert)

    def _configure_connection(self, **kwargs):
        # turn on foreign keys
        # http://www.sqlite.org/foreignkeys.html
        self._raw('PRAGMA foreign_keys = ON', ignore_result=True, **kwargs)

        # by default we can read/write, so only bother to run this query if we need to
        # make the connection actually readonly
        if self.config.readonly:
            self._readonly(self.config.readonly, **kwargs)

    def _get_connection(self):
        return self._connection

    def _close(self):
        self._connection.close()
        self._connection = None

    def _readonly(self, readonly, **kwargs):
        self._raw(
            # https://stackoverflow.com/a/49630725/5006
            'PRAGMA query_only = {}'.format("ON" if readonly else "OFF"),
            ignore_result=True,
            **kwargs
        )

    def _get_tables(self, table_name, **kwargs):
        query_str = 'SELECT tbl_name FROM sqlite_master WHERE type = ?'
        query_args = ['table']

        if table_name:
            query_str += ' AND name = ?'
            query_args.append(str(table_name))

        ret = self._raw(query_str, *query_args, **kwargs)
        return [r['tbl_name'] for r in ret]

    def _get_indexes(self, schema, **kwargs):
        """return all the indexes for the given schema"""
        # http://www.sqlite.org/pragma.html#schema
        # http://www.mail-archive.com/sqlite-users@sqlite.org/msg22055.html
        # http://stackoverflow.com/questions/604939/
        ret = {}
        rs = self._raw('PRAGMA index_list({})'.format(self.render_table_name_sql(schema)), **kwargs)
        if rs:
            for r in rs:
                iname = r['name']
                ret.setdefault(iname, [])
                indexes = self._raw('PRAGMA index_info({})'.format(r['name']), **kwargs)
                for idict in indexes:
                    ret[iname].append(idict['name'])

        return ret

    def _delete_tables(self, **kwargs):
        self._raw('PRAGMA foreign_keys = OFF', ignore_result=True, **kwargs);
        ret = super()._delete_tables(**kwargs)
        self._raw('PRAGMA foreign_keys = ON', ignore_result=True, **kwargs);
        return ret

    def _delete_table(self, schema, **kwargs):
        """
        https://www.sqlite.org/lang_droptable.html
        """
        #query_str = 'DROP TABLE IF EXISTS {}'.format(str(schema))
        query_str = "DROP TABLE IF EXISTS {}".format(self.render_table_name_sql(schema))
        self._raw(query_str, ignore_result=True, **kwargs)

    def _inserts(self, schema, field_names, field_values, **kwargs):
        """
        https://docs.python.org/3/library/sqlite3.html#sqlite3.Cursor.executemany
        """
        query_str = self.render_inserts_sql(schema, field_names, **kwargs)
        with self.connection(**kwargs) as connection:
            cur = connection.cursor()
            cur.executemany(
                query_str,
                field_values,
            )

    def create_error(self, e, **kwargs):
        kwargs.setdefault("error_module", sqlite3)
        if isinstance(e, sqlite3.OperationalError):
            e_msg = str(e)
            if "no such column" in e_msg or "has no column" in e_msg:
                #INSERT: "table yscrmiklbgdtx has no column named che"
                #SELECT: "no such column: che"
                e = FieldError(e)

            elif "no such table" in e_msg:
                e = TableError(e)

            elif "UNIQUE" in e_msg:
                e = UniqueError(e)

            elif "database is locked":
                e = CloseError(e)

            else:
                e = super().create_error(e, **kwargs)

        elif isinstance(e, sqlite3.IntegrityError):
            e = UniqueError(e)

        elif isinstance(e, sqlite3.ProgrammingError):
            e_msg = str(e)
            if "closed database" in e_msg:
                e = CloseError(e)

            elif "Incorrect number of bindings supplied" in e_msg:
                e = PlaceholderError(e, *kwargs.get("error_args", []))

            else:
                e = super().create_error(e, **kwargs)

        elif isinstance(e, sqlite3.InterfaceError):
            e_msg = str(e)
            if "Error binding parameter" in e_msg and "error_args" in kwargs:
                error_args = kwargs["error_args"]
                ms = re.search(r"parameter\s(\d+)", e_msg)
                index = int(ms.group(1))
                value = error_args[1][index]
                e = PlaceholderError(
                    e,
                    *error_args,
                    message=f"Placeholder {index} of query has unexpected type {type(value)}",
                )

            else:
                e = super().create_error(e, **kwargs)

        else:
            e = super().create_error(e, **kwargs)

        return e

    def _get_fields(self, table_name, **kwargs):
        """return all the fields for the given table"""
        ret = {}
        query_str = 'PRAGMA table_info({})'.format(self.render_table_name_sql(table_name))
        fields = self._raw(query_str, **kwargs)
        #pout.v([dict(d) for d in fields])

        query_str = 'PRAGMA foreign_key_list({})'.format(self.render_table_name_sql(table_name))
        fks = {f["from"]: f for f in self._raw(query_str, **kwargs)}
        #pout.v([dict(d) for d in fks.values()])

        pg_types = {
            "INTEGER": int,
            "BIGINT": long,
            "DOUBLE PRECISION": float,
            "FLOAT": float,
            "REAL": float,
            "NUMERIC": decimal.Decimal,
            "BOOLEAN": bool,
            "DATE": datetime.date,
            "TIMESTAMP": datetime.datetime,
            "CHARACTER": str,
            "VARCHAR": str,
            "TEXT": str,
            "BLOB": bytearray,
        }

        for field_type, adapter in sqlite3.adapters.items():
            if adapter_class := getattr(adapter, "__self__", None):
                if adapter_class.FIELD_TYPE not in pg_types:
                    pg_types[adapter_class.FIELD_TYPE] = field_type[0]

        # the rows we can set: field_type, name, field_required, min_size, max_size,
        #   size, unique, pk, <foreign key info>
        for row in fields:
            field = {
                "name": row["name"],
                "field_required": bool(row["notnull"]) or bool(row["pk"]),
                "pk": bool(row["pk"]),
            }

            for tname, ty in pg_types.items():
                if row["type"].startswith(tname):
                    field["field_type"] = ty
                    break

            if row["name"] in fks:
                field["schema_table_name"] = fks[row["name"]]["table"]
                field["ref_table_name"] = fks[row["name"]]["table"]

            ret[field["name"]] = field

        return ret

    def render_date_field_sql(self, field_name, field_kwargs, symbol):
        """
        allow extracting information from date

        http://www.sqlite.org/lang_datefunc.html
        """
        fstrs = []
        k_opts = {
            'day': "CAST(strftime('%d', {}) AS integer)",
            'hour': "CAST(strftime('%H', {}) AS integer)",
            'doy': "CAST(strftime('%j', {}) AS integer)", # day of year
            'julian_day': "strftime('%J', {})", # YYYY-MM-DD
            'month': "CAST(strftime('%m', {}) AS integer)",
            'minute': "CAST(strftime('%M', {}) AS integer)",
            'dow': "CAST(strftime('%w', {}) AS integer)", # day of week 0 = sunday
            'week': "CAST(strftime('%W', {}) AS integer)",
            'year': "CAST(strftime('%Y', {}) AS integer)"
        }

        for k, v in field_kwargs.items():
            fstrs.append([k_opts[k].format(self.render_field_name_sql(field_name)), self.PLACEHOLDER, v])

        return fstrs

    def render_sort_field_sql(self, field_name, field_vals, sort_dir_str):
        """
        allow sorting by a set of values

        http://stackoverflow.com/questions/3303851/sqlite-and-custom-order-by
        """
        fvi = None
        if sort_dir_str == 'ASC':
            fvi = (t for t in enumerate(field_vals)) 

        else:
            fvi = (t for t in enumerate(reversed(field_vals))) 

        query_sort_str = ['  CASE {}'.format(self.render_field_name_sql(field_name))]
        query_args = []
        for i, v in fvi:
            query_sort_str.append('    WHEN {} THEN {}'.format(self.PLACEHOLDER, i))
            query_args.append(v)

        query_sort_str.append('  END')
        query_sort_str = "\n".join(query_sort_str)
        return query_sort_str, query_args

    def render_datatype_int_sql(self, field_name, field, **kwargs):
        if field.is_auto():
            field_type = 'INTEGER'

        else:
            # we could break these up into tiny, small, and big but it
            # doesn't really matter so we're not bothering
            # https://www.sqlite.org/datatype3.html
            size = field.size_info()["size"]

            if size <= 9223372036854775807:
                field_type = 'INTEGER'

            else:
                field_type = NumericType.FIELD_TYPE

        return field_type

    def render_datatype_str_sql(self, field_name, field, **kwargs):
            field_type = super().render_datatype_str_sql(field_name, field, **kwargs)

            fo = field.interface_options
            if fo.get('ignore_case', False):
                field_type += ' COLLATE NOCASE'

            return field_type

    def render_datatype_datetime_sql(self, field_name, field, **kwargs):
        return DatetimeType.FIELD_TYPE

    def render_datatype_dict_sql(self, field_name, field, **kwargs):
        return DictType.FIELD_TYPE

    def render_datatype_uuid_sql(self, field_name, field, **kwargs):
        return self.render_datatype_str_sql(field_name, field)




# -*- coding: utf-8 -*-
from __future__ import unicode_literals, division, print_function, absolute_import

from datatypes import Enum

from ..config import (
    Field as BaseField,
)


class Field(BaseField):
    """Adds support for Enum as the field type. This allows you to set an enum string
    value, integer value, or Enum instance to the Orm or the Query and it will 
    be inserted into the db as an integer"""
    def is_enum(self):
        """Return True if the field type is an Enum"""
        try:
            ret = issubclass(self.original_type, Enum)
        except TypeError:
            ret = False
        return ret

    def fset(self, orm, val):
        if val is not None and self.is_enum():
            val = self.original_type.find_value(val)
        else:
            val = super(Field, self).fset(orm, val)
        return val

    def iquery(self, query, val):
        if val is not None and self.is_enum():
            val = self.original_type.find_value(val)
        else:
            val = super(Field, self).iquery(query, val)
        return val

    def set_type(self, field_type):
        super(Field, self).set_type(field_type)
        if self.is_enum():
            self.serializer = ""
            self._interface_type = int




# -*- coding: utf-8 -*-
from __future__ import unicode_literals, division, print_function, absolute_import
import os
import logging

from .config import (
    DsnConnection,
    Schema,
    Field,
    Index
)
from .query import Query, Iterator
from .model import Orm
from .interface import (
    get_interface,
    set_interface,
    get_interfaces,
    configure,
    configure_environ
)
from .exception import (
    Error,
    InterfaceError,
    UniqueError,
    TableError,
    FieldError,
    UniqueError,
    CloseError,
)

from . import utils


__version__ = '4.7.0'


def transaction(connection_name="", **kwargs):
    """Create a transaction 

    Sometimes you just need to batch a whole bunch of operation across a whole bunch
    of different models, this allows you to create a transaction outside of any
    of the models so you can do that, it will yield a connection you can then pass
    into the Orm/Query methods

    :Example:
        with prom.transaction(prefix="batch") as conn:
            o = FooOrm(foo=1)
            o.save(connection=conn)

    :param connection_name: str, the connection name corresponding to the anchor
        of the DSN, defaults to the default "" connection
    :param **kwargs: passed through to the Interface.transaction context manager
        * prefix: str, the name of the transaction you want to use
        * nest: bool, True if you want nested transactions to be created, False
            to ignore nested transactions
    :returns: Connection instance
    """
    kwargs.setdefault("nest", False)
    kwargs.setdefault("prefix", f"prom_{connection_name}_tx")
    return get_interface(connection_name).transaction(**kwargs)




# -*- coding: utf-8 -*-
from __future__ import unicode_literals, division, print_function, absolute_import
import decimal
import datetime
import inspect
import re
import base64
import json
import logging
import uuid

import dsnparse
from datatypes import (
    Datetime,
    property as cachedproperty,
)

from .compat import *
from . import utils


logger = logging.getLogger(__name__)


class Connection(object):
    """
    set the paramaters you want to use to connect to an interface

    https://github.com/Jaymon/Mingo/blob/master/Mingo/MingoConfig.php
    """
    name = ""
    """string -- the name of this connection (eg, Postgres, or SQLite)"""

    interface_name = ""
    """string -- full Interface class name -- the interface the connection should use to talk with the db"""

    host = ""
    """the hostname"""

    database = ""
    """the db name to use, in postgres, this is the database name"""

    port = 0
    """the host port"""

    username = ""
    """the username to use to to connect to the db"""

    password = ""
    """the password for the username"""

    readonly = False
    """Set to true to make the connection readonly"""

    options = None
    """any other db options, these can be interface implementation specific"""

    @property
    def interface(self):
        """Return a new Interface instance using this configuration"""
        interface_class = self.interface_class
        return interface_class(self)

    def __init__(self, fields=None, **fields_kwargs):
        """
        set all the values by passing them into this constructor, any unrecognized kwargs get put into .options

        :example:
            c = Connection(
                host="127.0.0.1",
                database="dbname",
                port=5000,
                some_random_thing="foo"
            )

            print(c.port) # 5000
            print(c.options) # {"some_random_thing": "foo"}
        """
        kwargs = utils.make_dict(fields, fields_kwargs)
        if "interface_name" not in kwargs:
            raise ValueError("no interface_name passed into Connection")

        self.options = {}

        for key, val in kwargs.items():
            if hasattr(self, key):
                setattr(self, key, val)
            else:
                self.options[key] = val

        interface_module, interface_class = utils.get_objects(self.interface_name)
        self.interface_class = interface_class
        interface_class.configure(self)


class DsnConnection(Connection):
    """
    sets up a connection configuration using a "prom dsn" or connection string

    the prom dsn is in the form:

        <full.python.path.InterfaceClass>://<username>:<password>@<host>:<port>/<database>?<options=val&query=string>#<name>

    This is useful to allow connections coming in through environment variables as described in :
        http://www.12factor.net/backing-services

    It tooks its inspiration from this project:
        https://github.com/kennethreitz/dj-database-url

    http://en.wikipedia.org/wiki/Connection_string
    http://en.wikipedia.org/wiki/Data_source_name
    """
    dsn = ""
    """holds the raw dsn string that was parsed"""

    def __init__(self, dsn):
        # get the scheme, which is actually our interface_name
        kwargs = self.parse(dsn)
        super().__init__(**kwargs)

    def parse(self, dsn):
        parser = dsnparse.parse(dsn)

        d = parser.fields
        d["dsn"] = parser.parser.dsn

        # remap certain values
        d["name"] = d.pop("fragment")
        d["interface_name"] = self.normalize_scheme(d.pop("scheme"))
        d["database"] = d.pop("path")
        d["options"] = d.pop("query_params", {})
        d["host"] = d.pop("hostname")
        d["readonly"] = bool(d["options"].pop("readonly", self.readonly))

        # get rid of certain values
        d.pop("params", None)
        return d

    def normalize_scheme(self, v):
        ret = v
        d = {
            "prom.interface.sqlite.SQLite": set(["sqlite"]),
            "prom.interface.postgres.PostgreSQL": set(["postgres", "postgresql", "psql"])
        }

        kv = v.lower()
        for interface_name, vals in d.items():
            if kv in vals:
                ret = interface_name
                break

        return ret


class Schema(object):
    """
    handles all table schema definition

    the table schema definition includes the table name, the fields the table has, and
    the indexes that are on the table
    """
    instances = {}
    """class variable, holds different schema instances for various orms"""

    table_name = ""
    """string -- set the table name for this schema instance"""

    fields = None
    """dict -- all the fields this schema instance will use"""

    indexes = None
    """dict -- all the indexes this schema will have"""

    lookup = None
    """dict -- field information lookup table, basically an internal cache"""

    @property
    def normal_fields(self):
        """fields that aren't magic (eg, don't start with an underscore)"""
        return {f:v for f, v in self.fields.items() if not f.startswith('_')}

    @property
    def required_fields(self):
        """The normal required fields (eg, no magic fields like _id are included)"""
        return {f:v for f, v in self.normal_fields.items() if v.required}

    @property
    def ref_fields(self):
        """Return FK reference fields"""
        return {f:v for f, v in self.normal_fields.items() if v.is_ref()}

    @property
    def magic_fields(self):
        """the magic fields for the schema, magic fields start with an underscore"""
        return {f:v for f, v in self.fields.items() if f.startswith('_')}

    @property
    def pk_fields(self):
        pk_fields = {}
        for pk_name in self.pk_names:
            pk_fields[pk_name] = self.fields[pk_name]
        return pk_fields

    @property
    def pk_name(self):
        """returns the primary key name for this schema"""
        try:
            pk_field = self.__getattr__("pk")
            pk_name = pk_field.name
        except AttributeError:
            pk_name = None
        return pk_name

    @property
    def pk_names(self):
        """Returns all the field names comprising the primary key as a list"""
        pk_name = self.pk_name
        return [pk_name] if pk_name else []

    @property
    def schemas(self):
        """Find and return all the schemas that are needed for this schema to 
        install successfully

        Another way to put this is all the schemas this Schema touches

        :returns: list, a list of Schema instances, self will be at the end
        """
        schemas = []
        for f in self.fields.values():
            if s := f.schema:
                schemas.extend(s.schemas)
        schemas.append(self)

        return schemas

    @classmethod
    def get_instance(cls, orm_class, **kwargs):
        """return a Schema singleton instance for the given orm_class

        if there isn't already an instance in cache then a new instance will be
        created. If a Schema instance is already in cache then it will be returned

        :param orm_class: Orm, the class to create the schema for
        :returns: Schema
        """
        table_name = orm_class.table_name
        if table_name in cls.instances:
            s = cls.instances[table_name]

        else:
            s = cls(table_name)
            s.orm_class = orm_class

            seen_properties = {}
            for klass in inspect.getmro(orm_class)[:-1]:
                for k, v in vars(klass).items():
                    k = String(k)
                    field = None

                    if isinstance(v, Field):
                        field = v

                    elif isinstance(v, Index):
                        s.set_index(k, v)
                        seen_properties[k] = v

                    elif isinstance(v, type) and issubclass(v, Field):
                        # We've defined a Field class inline of the Orm, so
                        # we want to instantiate it and set it in all the places
                        field = v.get_instance()

                    else:
                        if v is None:
                            seen_properties[k] = v

                    if field:
                        if k in seen_properties:
                            if seen_properties[k] is None:
                                for fn in field.names:
                                    s.lookup["field_names_deleted"][fn] = k

                        else:
                            s.set_field(k, field)
                            seen_properties[k] = field

            cls.instances[table_name] = s

        return s

    def __init__(self, table_name, **fields_or_indexes):
        """Create an instance

        every Orm should have a .schema attribute that is an instance of this class

        :example:
            schema = Schema(
                "table_name"
                field1=Field(int, True),
                field2=Field(str),
                index_fields=Index("field1", "field2")
            )

        :param table_name: string, the table name
        :param **fields_or_indexes: a dict of field name or index keys with tuple values,
            see __getattr__ for more details
        """
        self.fields = {}
        self.indexes = {}
        self.table_name = String(table_name)
        self.orm_class = None
        self.lookup = {
            "field_names": {},
            # holds parent's Field.names that have been set to None in children
            # classes. Basically, if a parent class sets foo and then a child class
            # later on sets foo = None then foo.names will be here
            "field_names_deleted": {},
        }

        for name, val in fields_or_indexes.items():
            self.set(name, val)

    def __str__(self):
        return self.table_name

    def __iter__(self):
        """iterate through all the fields defined in this schema

        :returns: tuple, (name, Field)
        """
        return (f for f in self.fields.items())

    def set(self, name, val):
        if isinstance(val, Field):
            self.set_field(name, val)

        elif isinstance(val, Index):
            self.set_index(name, val)

        else:
            raise TypeError("Not a Field or Index instance")

    def __getattr__(self, name):
        """
        return the Field instance for the given name

        return -- string -- the string value of the attribute name, eg, self.foo returns "foo"
        """
        try:
            return self.lookup["field_names"][name]

        except KeyError:
            raise AttributeError("No {} field in schema {}".format(name, self.table_name))

    def has_field(self, field_name):
        """Return True if schema contains field_name"""
        return field_name in self.lookup["field_names"]

    def __contains__(self, field_name):
        return self.has_field(field_name)

    def set_field(self, field_name, field):
        if not field_name:
            raise ValueError("field_name is empty")
        if field_name in self.fields:
            raise ValueError(f"{field_name} already exists and cannot be changed")
        if not isinstance(field, Field):
            raise ValueError(f"{type(field)} is not a Field instance")

        field.__set_name__(self.orm_class, field_name)

        if field.unique:
            self.set_index(field_name, Index(field_name, unique=True))

        if field.index:
            self.set_index(field_name, Index(field_name))

        self.fields[field_name] = field

        for fn in field.names:
            if fn in self.lookup["field_names"] and fn in field.aliases:
                self.lookup["field_names"].pop(fn)
                logger.warning(
                    " ".join([
                        "{} ignored alias {} for {} because it is".format(
                            self,
                            fn,
                            field.name,
                        ),
                        "used by another field",
                    ])
                )

            else:
                self.lookup["field_names"][fn] = field

        return self

    def set_index(self, index_name, index):
        """Add an index to the schema

        for the most part, you will use the __getattr__ method of adding indexes for a more fluid interface,
        but you can use this if you want to get closer to the bare metal

        index_name -- string -- the name of the index
        index -- Index() -- an Index instance
        """
        if not index_name:
            raise ValueError("index_name must have a value")
        if index_name in self.indexes:
            raise ValueError("index_name {} has already been defined on {}".format(
                index_name, str(self.indexes[index_name].field_names)
            ))
        if not isinstance(index, Index):
            raise ValueError(f"{type(index)} is not an Index instance")

        index.name = index_name
        index.orm_class = self.orm_class

        self.indexes[index_name] = index
        return self

    def field_name(self, field_name, *default):
        """Get the canonical field name of field_name

        most of the time, the field name of field_name will just be field_name,
        but this checks the configured aliases to return the canonical name

        :param field_name: str, the field_name you want the canonical field name for
        :param *default: mixed, if present this will be returned instead of AttributeError
            raised if field_name doesn't exist
        """
        try:
            return self.__getattr__(field_name).name

        except AttributeError:
            if default:
                return default[0]

            else:
                if field_name in self.lookup["field_names_deleted"]:
                    return self.lookup["field_names_deleted"][field_name]

                else:
                    raise

    def field_model_name(self, field_name):
        """Check field_name against all the field's ref model names to see if there
        is a match, this is separate from .field_name because there are times when
        this behavior might not be desirable

        :param field_name: str, the field/model name you want the canonical field
            name for
        :returns: str, the canonical field name
        """
        for fn, field in self.fields.items():
            if field_name in field.ref_names:
                return fn

        raise AttributeError(field_name)

    def create_orm(self, orm_class=None):
        """If you have a schema but don't have an Orm for it, you can call this method
        and have an orm_class created that will have the fields and table_name of this
        schema

        :param orm_class: Orm, if you want your generated class to have a certain
            parent class you can pass in the parent class you want
        :returns: Orm, your child orm populated with information from this schema
        """
        if not orm_class:
            from .model import Orm # avoid circular dependency
            orm_class = Orm

        child_class = type(
            String(self.table_name),
            (orm_class,),
            {"table_name": self.table_name, "schema": self}
        )

        # clear all the Field and Index properties in anticipation of adding the
        # fields from the Schema
        seen_properties = set()
        for klass in inspect.getmro(child_class):
            for k, v in vars(klass).items():
                if k not in seen_properties:
                    if isinstance(v, (Field, Index)):
                        #setattr(child_class, k, property(lambda: AttributeError))
                        setattr(child_class, k, None)
                        #delattr(child_class, k)

                    elif isinstance(v, type) and issubclass(v, Field):
                        setattr(child_class, k, None)

                    seen_properties.add(k)

        for field_name, field in self:
            setattr(child_class, field_name, field)

        return child_class


class Index(object):
    """Each index on the table is configured using this class"""

    def __init__(self, *field_names, **options):
        """
        initialize an index

        index_fields -- list -- the string field_names this index will index on, fields have to be already added
            to this schema index
        **options -- dict --
            unique -- boolean -- True if the index should be unique, false otherwise
        """
        if not field_names:
            raise ValueError("field_names list is empty")

        self.name = ""
        self.field_names = list(map(String, field_names))
        self.options = options
        self.unique = options.get("unique", False)


class FieldMeta(type):
    """Allows a class definition to be a descriptor also

    I don't love this solution, but I like the syntax of just being able to define
    a subclass inside an Orm and have it work. The problem is an embedded class isn't,
    by default, a descriptor instance, so it wasn't calling fset, fget, and fdel.

    This makes it so an embedded field class will be treated as a descriptor. It
    uses an embedded .instance property to actually perform the operations
    """
    def __set_name__(cls, orm_class, name):
        """I'm not quite sure why this works, but it does. Basically, when you define
        the field as an embedded class in the orm, this method will get called
        when the class is being parsed/loaded the first time.

        So this method gets called and we use cls to create a Field instance and
        we then set our newly created instance onto orm_class.name. So after this
        method is called, orm_class.name will be a Field instance instead of a
        Field subclass
        """
        logger.debug(f"Creating field descriptor {orm_class.__name__}.{name} with class {cls.__name__}")

        instance = cls(cls.type, cls.required, cls.options)
        instance.__set_name__(orm_class, name)
        setattr(orm_class, name, instance)


class Field(object, metaclass=FieldMeta):
    """Each column in the database is configured using this class

    You can set some getters and setters on this object in order to fine tune
    control over how a field in the db is set and fetched from the instance and
    also how it is set and fetched from the interface

    to customize the field on set and get from the db, you can use the decorators,
    which are classmethods:

    :Example:
        foo = Field(str, True)

        @foo.igetter
        def foo(self, val):
            # do custom things when pulling foo from the database
            return val

        @foo.isetter
        def foo(self, val):
            # do custom things when putting foo into the database
            return val

    There are also fget/fset/fdel methods that can be set to customize behavior
    on when a value is set on a particular orm instance, so if you wanted to make sure
    that bar was always an int when it is set, you could:

    :Example:
        bar = Field(int, True)

        @bar.fsetter
        def bar(self, val):
            # do custom things
            return int(val)

        @bar.fgetter
        def bar(self, val):
            return int(val)

        @bar.jsonabler
        def bar(self, name, val):
            # convert val to something json safe, you can also change the name
            return name, val

    NOTE -- all these get/set/delete methods are different than traditional python getters
    and setters because they always need to return a value and they always take in the
    object instance manipulating them and the value.

    NOTE -- Foreign key's can be passed orm instances from other classes because
    those classes will call the FK's field methods when getting/setting the field

    You can also configure the Field using a class:

    :Example:
        class bar(Field):
            type = int
            required = True
            options = {
                "default": 0,
                "unique": True,
                "max_size": 512,
            }

            def fget(self, orm, v):
                print("fget")
                return v

            def iget(self, orm, v):
                print("iget")
                return v

            def fset(self, orm, v):
                print("fset")
                return v

            def fdel(self, orm, v):
                print("fdel")
                return v

            def iquery(self, query, v):
                print("iquery")
                return v

            def jsonable(self, orm, field_name, v):
                print("jsonable")
                return field_name, v

    https://docs.python.org/2/howto/descriptor.html
    """
    required = False
    """True if this field is required"""

    options = None
    """In the instance, this will be a dict of key/val pairs containing extra information
    about the field"""

    default = None
    """Default value for this field"""

    help = ""
    """The description/help message for this field"""

    name = ""
    """The field name"""

    choices = None
    """A set of values that this field can be, if set then no other values can be set"""

    unique = False
    """True if this field is unique indexed"""

    index = False
    """True if this field is indexed"""

    orm_class = None
    """Holds the model class this field is defined on"""

    @cachedproperty(cached="_schema")
    def schema(self):
        """return the schema instance if this field is a reference to another table

        see .set_type() for an explanation on why we defer figuring this out until now
        """
        field_type = self.original_type
        module, klass = utils.get_objects(field_type)
        schema = klass.schema
        if not schema:
            raise ValueError("Field type {} is not an Orm class".format(field_type))

        return schema

    @property
    def type(self):
        """alias of the interface type, this is really here so you can set it to
        a value if you are inline defining a field, this will get passed to the
        __init__ method and then be used to set ._interface_type"""
        return self.interface_type

    @property
    def interface_type(self):
        """Returns the type that will be used in the interface to create the table

        see .set_type() for an explanation on why we defer this until here
        """
        if self._interface_type is None:
            if self.is_serialized():
                self._interface_type = str

            else:
                s = self.schema
                if s:
                    self._interface_type = s.pk.type

                else:
                    self._interface_type = self.original_type

        return self._interface_type

    @property
    def ref(self):
        """Returns the FK reference orm class"""
        schema = self.schema
        return schema.orm_class if schema else None

    @property
    def ref_class(self):
        return self.ref

    @property
    def names(self):
        names = set()
        if self.name:
            names.add(self.name)

        names.update(self.aliases)

        return names

    @property
    def aliases(self):
        names = set()
        names.update(self.options.get("names", []))
        names.update(self.options.get("aliases", []))

        if alias := self.options.get("alias", ""):
            names.add(alias)

        if self.is_pk():
            names.add("pk")
            if self.orm_class:
                model_name = self.orm_class.model_name
                names.add(f"{model_name}_id")
                names.add(f"{model_name}_pk")

        return names

    @property
    def ref_names(self):
        names = set()
        if ref_class := self.ref:
            names.add(ref_class.model_name)
            names.add(ref_class.models_name)
        return names

    @property
    def interface_options(self):
        """When the interface is doing stuff with the field it might call this
        method to make sure it has all the options it needs to use the field
        """
        if self.is_ref():
            options = self.schema.pk.options
            options.update(self.options)
        else:
            options = self.options
        return options

    def __init__(self, field_type, field_required=False, field_options=None, **field_options_kwargs):
        """
        create a field

        :param field_type: type, the python type of the field, so for a string
            you would pass str, integer: int, boolean: bool, float: float
        :param field_required: boolean, true if this field has to be there to insert
        :param field_options: dict, everything else in key: val notation. Current options:
            * size: int -- the size you want the string to be, or the int to be
            * min_size: int -- the minimum size
            * max_size: int -- if you want a varchar, set this
            * unique: bool -- True to set a unique index on this field, this is just for convenience and is
                equal to self.set_index(field_name, [field_name], unique=True). this is a convenience option
                to set a unique index on the field without having to add a separate index statement
            * ignore_case: bool -- True to ignore case if this field is used in indexes
            * default: mixed -- defaults to None, can be anything the db can support
            * jsonable_name: str, the name of the field when Orm.jsonable() is called
            * jsonable_field: bool, if False then this field won't be part of
                Orm.jsonable() output
            * jsonable: str|callable|bool,
                - str, will be set to jsonable_name
                - callable, will be used as self.jsonable
                - bool, will set jsonable_field to False
            * empty: bool, (default is True), set to False if the value cannot be
                empty when being sent to the db (empty is None, "", 0, False)
            * pk: bool, True to make this field the primary key
            * auto: bool, True to tag this field as an auto-generated field. Used
                with an int to set it to an auto-increment, use it with UUID to
                have uuids auto generated
            * autopk: bool, shortcut for setting pk=True, auto=True
        :param **field_options_kwargs: dict, will be combined with field_options
        """
        field_options = utils.make_dict(field_options, field_options_kwargs)

        d = self.get_size(field_options)
        field_options.update(d)

        name = field_options.pop("name", "")
        orm_class = field_options.pop("orm_class", None)

        choices = field_options.pop("choices", set())
        if choices or not self.choices:
            self.choices = choices

        if jsonable := field_options.pop("jsonable", None):
            if isinstance(jsonable, str):
                field_options.setdefault("jsonable_name", jsonable)

            elif isinstance(jsonable, bool):
                field_options.setdefault("jsonable_field", jsonable)

            else:
                field_options["jsonable"] = jsonable

        if autopk := field_options.pop("autopk", False):
            field_options["pk"] = autopk
            field_options["auto"] = autopk

        for k in list(field_options.keys()):
            if hasattr(self, k):
                setattr(self, k, field_options.pop(k))

        self.options = field_options
        self.required = field_required or self.is_pk()

        self.set_type(field_type)

        # this class is being created manually so mimic what the python parser
        # would do
        if name and orm_class:
            self.__set_name__(orm_class, name)

        elif orm_class:
            self.orm_class = orm_class

        elif name:
            self.__set_name__(orm_class, name)

    def __set_name__(self, orm_class, name):
        """This is called right after __init__

        https://docs.python.org/3/howto/descriptor.html#customized-names

        This is only called when an instance is created while a class is being
        parsed/created, so if you just created an instance of Field you would
        need to call this method manually

        :param orm_class: type, the orm class this Field will belong to
        :param name: str, the field's public name on the orm class
        """
        self.orm_class = orm_class
        self.name = name

        # the field name this descriptor will use to set the value onto the orm
        # instance
        self.orm_field_name = f"_{name}"

        # we keep a hash of the field's value when it was pulled from the
        # interface (see .iget) so we know if the field has been modified
        self.orm_interface_hash = f"_interface_{name}_hash"

        if orm_class:
            logger.debug(
                f"Field descriptor {orm_class.__name__}.{name} created with internal name {self.orm_field_name}"
            )

    def get_size(self, field_options):
        """Figure out if this field has any size information"""
        d = {}

        min_size = field_options.pop("min_size", -1)
        max_size = field_options.pop("max_size", -1)
        size = field_options.pop("size", None)

        if size is not None:
            d['size'] = size

        else:
            if min_size > 0 and max_size < 0:
                d['min_size'] = min_size

            elif min_size < 0 and max_size > 0:
                d['max_size'] = max_size

            elif min_size > 0 and max_size > 0:
                d['min_size'] = min_size
                d['max_size'] = max_size

        return d

    def size_info(self):
        """Figure out actual sizing information with all the ways we can calculate
        size now

        :returns: dict, will always have "has_size" and "has_precision" keys
            * if "has_size" key is True then "size" key will exist
            * if "has_precision" is True then "precision" and "scale" keys will exist
            * if "bounds" key exists it will be a tuple (min_size, max_size)
        """
        ret = {
            "has_size": False,
            "has_precision": False,
            "original": {},
        }
        options = self.interface_options

        if "size" in options:
            ret["size"] = options["size"]
            ret["original"]["size"] = options["size"]
            ret["has_size"] = True
            ret["bounds"] = (ret["size"], ret["size"])

        if "max_size" in options:
            ret["size"] = max(options["max_size"], ret.get("size", 0))
            ret["original"]["max_size"] = options["max_size"]
            ret["has_size"] = True
            ret["bounds"] = (ret.get("size", 0), ret["size"])

        if "min_size" in options:
            ret["size"] = max(options["min_size"], ret.get("size", 0))
            ret["original"]["min_size"] = options["min_size"]
            ret["has_size"] = True
            ret["bounds"] = (options["min_size"], ret["size"])

        if ret["has_size"]:
            # if size is like 15.6 then that would be considered 21
            # precision with a scale of 6 (ie, you can have 15 digits before
            # the decimal point and 6 after)
            parts = str(ret["size"]).split(".")
            if len(parts) > 1:
                ret["scale"] = int(parts[1])
                ret["precision"] = int(parts[0]) + ret["scale"]

                # the set size was actually precision and scale so we don't have
                # sizing information
                ret["has_precision"] = True
                ret.pop("size")

        if "precision" in options:
            # So the number 23.5141 has a precision of 6 and a scale of 4
            precision = options["precision"]
            ret["precision"] = int(precision)
            ret["original"]["precision"] = precision
            ret["has_precision"] = True
            ret["scale"] = 0

            if "scale" in options:
                scale = options["scale"]
                ret["scale"] = int(scale)
                ret["original"]["scale"] = scale

        interface_type = self.interface_type
        if issubclass(interface_type, (int, float, decimal.Decimal)):
            if "precision" in ret and "size" not in ret:
                if ret["scale"]:
                    ret["size"] = float("{}.{}".format(
                        "9" * (ret["precision"] - ret["scale"]),
                        "9" * ret["scale"],
                    ))

                else:
                    ret["size"] = int("9" * ret["precision"])

            elif "precision" not in ret and "size" not in ret:
                # this is 32bit, it might be worth setting defined size to 64bit
                ret["size"] = 2147483647
                ret["precision"] = len(str(ret["size"]))

            elif "precision" not in ret and "size" in ret:
                ret["precision"] = len(str(ret["size"]))

        elif issubclass(interface_type, (str, bytes, bytearray)):
            if ret["has_size"]:
                if not ret["has_precision"]:
                    ret["precision"] = ret["size"]
                    ret["scale"] = 0

            elif ret["has_precision"]:
                ret["size"] = ret["precision"]

        return ret

    def set_type(self, field_type):
        """Try to infer as much about the type as can be inferred at this moment

        Because the Field supports string classpaths (eg, "modname.Classname") we can't
        figure everything out in this method, so we figure out as much as we can
        and then defer everything else to the .interface_type and .schema properties, this
        allows the parser to hopefully finish loading the modules before we have to
        parse the classpath to find the foreign key schema

        :param field_type: mixed, the field type passed into __init__
        """
        std_types = (
            str,
            bool,
            int,
            float,
            bytes,
            bytearray,
            decimal.Decimal,
            datetime.datetime,
            datetime.date,
            uuid.UUID,
            dict,
        )

        json_types = ()

        pickle_types = (
            set,
            list,
        )

        self.original_type = field_type
        self.serializer = ""
        self._interface_type = None
        self._schema = None

        if isinstance(field_type, type):
            if issubclass(field_type, std_types):
                self._interface_type = field_type

            elif issubclass(field_type, json_types):
                self.serializer = self.options.pop("serializer", "json")
                self._interface_type = str

            elif issubclass(field_type, pickle_types):
                self.serializer = self.options.pop("serializer", "pickle")
                self._interface_type = bytes

            else:
                schema = getattr(field_type, "schema", None)
                if schema:
                    self._schema = schema

                else:
                    # We have just some random class that isn't an Orm
                    self.serializer = self.options.pop("serializer", "pickle")
                    self._interface_type = bytes

        elif isinstance(field_type, Schema):
            self._schema = field_type

        else:
            # check if field_type is a string classpath so we have to defer
            # setting the type
            if isinstance(field_type, basestring):
                # no ._schema property will make .schema treat .original_type as
                # a classpath
                del self._schema

            else:
                raise ValueError("Unknown field type {}".format(field_type))

    def is_pk(self):
        """return True if this field is a primary key"""
        return self.options.get("pk", False)

    def is_auto(self):
        """Return True if this field auto-generates its value somehow"""
        return self.options.get("auto", False)

    def is_ref(self):
        """return true if this field foreign key references the primary key of another orm"""
        return bool(self.schema)

    def is_required(self):
        """Return True if this field is required to save into the interface"""
        return self.required

    def is_serialized(self):
        """Return True if this field should be serialized"""
        return True if self.serializer else False

    def fget(self, orm, val):
        """Called anytime the field is accessed through the Orm (eg, Orm.foo)

        :param orm: Orm, the Orm instance the field is being accessed on
        :param val: mixed, the current value of the field
        :returns: mixed
        """
        logger.debug(f"fget {orm.__class__.__name__}.{self.name}")
        if self.is_ref():
            # Foreign Keys get passed through their Field methods
            val = self.schema.pk.fget(orm, val)
        return val

    def fgetter(self, v):
        """decorator for setting field's fget function"""
        self.fget = v
        return self

    def iget(self, orm, val):
        """Called anytime the field is being returned from the interface to the orm

        think of this as when the orm receives the field value from the interface

        :param orm: Orm, the Orm instance the field is being set on. This can be
            None if the select query had selected fields so a full orm instance isn't
            being returned but rather just the selected values
        :param val: mixed, the current value of the field
        :returns: mixed
        """
        logger.debug(f"iget {orm.__class__.__name__}.{self.name}")
        if self.is_ref():
            # Foreign Keys get passed through their Field methods
            val = self.schema.pk.iget(orm, val)

        else:
            if self.is_serialized():
                val = self.decode(val)

            if orm:
                orm.__dict__[self.orm_interface_hash] = self.hash(orm, val)

        return val

    def igetter(self, v):
        """decorator for the method called when a field is pulled from the database"""
        self.iget = v
        return self

    def fset(self, orm, val):
        """This is called on Orm instantiation and any time field is set (eg Orm.foo = ...)

        on Orm creation val will be None if the field wasn't passed to Orm.__init__ otherwise
        it will be the value passed into Orm.__init__

        :param orm: Orm, the Orm instance the field is being set on
        :param val: mixed, the current value of the field
        :returns: mixed
        """
        logger.debug(f"fset {orm.__class__.__name__}.{self.name}")
        if val is not None:
            if self.choices and val not in self.choices:
                raise ValueError("Value {} not in {} value choices".format(val, self.name))

            if regex := self.options.get("regex", ""):
                if not re.search(regex, val):
                    raise ValueError(f"regex failed for {orm.__class__.__name__}.{self.name}")

        if self.is_ref():
            # Foreign Keys get passed through their Field methods
            val = self.schema.pk.fset(orm, val)

        return val

    def fsetter(self, v):
        """decorator for setting field's fset function"""
        self.fset = v
        return self

    def iset(self, orm, val):
        """Called anytime the field is being fetched to send to the interface

        think of this as when the interface is going to get the field value or
        when the field is being sent to the db. Alot of the value checks like
        required and empty are in Orm.to_interface(), this just returns the value
        and nothing else

        :param orm: Orm
        :param val: Any, the current value of the field
        :returns: Any
        """
        logger.debug(f"iset {orm.__class__.__name__}.{self.name}")
        if self.is_ref():
            # Foreign Keys get passed through their Field methods
            val = self.schema.pk.iset(orm, val)

        else:
            if self.is_serialized():
                val = self.encode(val)

        return val

    def isetter(self, v):
        """decorator for setting field's fset function"""
        self.iset = v
        return self

    def fdel(self, orm, val):
        logger.debug(f"fdel {orm.__class__.__name__}.{self.name}")
        orm.__dict__.pop(self.orm_interface_hash, None)
        return None

    def fdeleter(self, v):
        """decorator for setting field's fdel function"""
        self.fdel = v
        return self

    def idel(self, orm, val):
        """Called when the field is being deleted from the db

        :param orm: Orm
        :param val: mixed, the current value of the field
        :returns: mixed
        """
        logger.debug(f"idel {orm.__class__.__name__}.{self.name}")
        orm.__dict__.pop(self.orm_interface_hash, None)
        return None if self.is_pk() else val

    def ideleter(self, v):
        """decorator for setting field's idel function"""
        self.idel = v
        return self

    def fdefault(self, orm, val):
        """On a new Orm instantiation, this will be called for each field and if
        val equals None then this will decide how to use self.default to set
        the default value of the field

        If you just want to set a default value you won't need to override this
        method because you can just pass default into the field instantiation
        and it will get automatically used in this method

        :param orm: Orm, the Orm instance being created
        :param val: mixed, the current value of the field (usually None)
        :returns: mixed
        """
        #pout.v("fdefault {}".format(self.name))
        ret = val
        if val is None:
            if callable(self.default):
                ret = self.default()

            elif self.default is None:
                ret = self.default

            elif isinstance(self.default, (dict, list, set)):
                ret = type(self.default)()

            else:
                ret = self.default

        return ret

    def fdefaulter(self, v):
        """decorator for returning the field's default value

        decorator for setting field's fdefault function"""
        self.fdefault = v
        return self

    def iquery(self, query, val):
        """This will be called when setting the field onto a query instance

        :example:
            o = Orm(foo=1)
            o.query.eq_foo(1) # iquery called here

        :param query: Query
        :param val: mixed, the fields value
        :returns: mixed
        """
        if self.is_ref():
            # Foreign Keys get passed through their Field methods
            val = self.schema.pk.iquery(query, val)
        return val

    def iquerier(self, v):
        """decorator for the method called when this field is used in a SELECT query"""
        self.iquery = v
        return self

    def jsonable(self, orm, name, val):
        """This is called in Orm.jsonable() to set the field name nad value

        :param orm: Orm, the instance currently calling jsonable
        :param name: str, the Orm field name that will be used if the jsonable_name
            option isn't set
        :param val: Any, the field's value
        :returns: tuple[str, Any], (name, val) where the name will be the jsonable
            field name and the value will be the jsonable value
        """
        if self.options.get("jsonable_field", True):
            logger.debug(f"jsonable {orm.__class__.__name__}.{self.name} -> {name}")

            if self.is_ref():
                # Foreign Keys get passed through their Field methods
                val = self.schema.pk.jsonable(orm, val)

            else:
                if val is None:
                    val = self.fdefault(orm, val)

                if val is not None:
                    format_str = ""
                    if isinstance(val, (datetime.datetime, datetime.date)):
                        val = Datetime(val).isoformat()

            return self.options.get("jsonable_name", name), val

        else:
            logger.debug(f"jsonable ignored for {orm.__class__.__name__}.{self.name}")
            return "", None

    def jsonabler(self, v):
        """Decorator for the method called for a field when an Orm's .jsonable method
        is called"""
        self.jsonable = v
        return self

    def modified(self, orm, val):
        """Returns True if val has been modified in orm

        :param orm: Orm
        :param val: mixed, the current value of the field
        :returns: bool, True if val is different than the interface val
        """
        if self.is_serialized():
            return True

        ret = True
        ihash = orm.__dict__.get(self.orm_interface_hash, None)
        if ihash:
            ret = self.hash(orm, val) != ihash

        else:
            ret = val is not None

        return ret

    def hash(self, orm, val):
        if self.is_serialized():
            return None
        elif isinstance(val, dict):
            return None
        return hash(val)

    def encode(self, val):
        if val is None: return val

        if self.serializer == "pickle":
            #return base64.b64encode(pickle.dumps(val, pickle.HIGHEST_PROTOCOL))
            return pickle.dumps(val, pickle.HIGHEST_PROTOCOL)

        elif self.serializer == "json":
            return json.dumps(val)

        else:
            raise ValueError("Unknown serializer {}".format(self.serializer))

    def decode(self, val):
        if val is None: return val

        if self.serializer == "pickle":
            #return pickle.loads(base64.b64decode(val))
            return pickle.loads(val)

        elif self.serializer == "json":
            return json.loads(val)

        else:
            raise ValueError("Unknown serializer {}".format(self.serializer))

    def fval(self, orm):
        """return the raw value that this property is holding internally for the orm instance"""
        try:
            val = orm.__dict__[self.orm_field_name]

        except KeyError as e:
            #raise AttributeError(str(e))
            val = None

        return val

    def __get__(self, orm, orm_class=None):
        """This is the wrapper that will actually be called when the field is
        fetched from the instance, this is a little different than Python's built-in
        @property fget method because it will pull the value from a shadow variable in
        the instance and then call fget"""
        if orm is None:
            # class is requesting this property, so return it
            return self

        raw_val = self.fval(orm)
        ret = self.fget(orm, raw_val)

        # we want to compensate for default values right here, so if the raw val
        # is None but the new val is not then we save the returned value, this
        # allows us to handle things like dict with no surprises
        if raw_val is None:
            if ret is not None:
                orm.__dict__[self.orm_field_name] = ret

        return ret

    def __set__(self, orm, val):
        """this is the wrapper that will actually be called when the field is
        set on the instance, your fset method must return the value you want set,
        this is different than Python's built-in @property setter because the
        fset method *NEEDS* to return something"""
        val = self.fset(orm, val)
        orm.__dict__[self.orm_field_name] = val

    def __delete__(self, orm):
        """the wrapper for when the field is deleted, for the most part the default
        fdel will almost never be messed with, this is different than Python's built-in
        @property deleter because the fdel method *NEEDS* to return something and it
        accepts the current value as an argument"""
        val = self.fdel(orm, self.fval(orm))
        orm.__dict__[self.orm_field_name] = val


class AutoUUID(Field):
    """an auto-generating UUID field, by default this will be set as primary key"""
    def __init__(self, **kwargs):
        kwargs.setdefault("pk", True)
        kwargs.setdefault("auto", True)
        kwargs.setdefault("size", 36)
        super().__init__(
            field_type=uuid.UUID,
            field_required=True,
            **kwargs
        )


class AutoIncrement(Field):
    """an auto-incrementing Serial field, by default this will be set as primary key"""
    def __init__(self, **kwargs):
        kwargs.setdefault("pk", True)
        kwargs.setdefault("auto", True)
        kwargs.setdefault("max_size", 9223372036854775807)
        super().__init__(
            field_type=int,
            field_required=True,
            **kwargs
        )


class AutoDatetime(Field):
    """A special field that will create a datetime according to triggers

    :Example:
        # datetime will be auto-populated on creation
        created = AutoDatetime(created=True)

        # datetime will be auto-populated on creation and update
        updated = AutoDatetime(updated=True)

    The triggers are:
        created: create a datetime in the field when the Orm is being created (added
            to the db for the first time)
        updated: update the datetime when the Orm is created and when it is updated
            in the db
    """
    def __init__(self, **kwargs):
        kwargs.setdefault("created", kwargs.get("create", True))
        kwargs.setdefault("updated", kwargs.get("update", False))
        kwargs.setdefault("auto", True)
        if kwargs["updated"]:
            kwargs["created"] = False

        super().__init__(
            field_type=datetime.datetime,
            field_required=True,
            **kwargs,
        )

    def iset(self, orm, val):
        if self.options.get("updated", False):
            return self.updated_iset(orm, val)

        elif self.options.get("created", False):
            return self.created_iset(orm, val)

        else:
            return super().iset(orm, val)

    def created_iset(self, orm, val):
        if not val and orm.is_insert():
            #val = datetime.datetime.utcnow()
            #val = datetime.datetime.now(datetime.timezone.utc)
            val = Datetime()
        return val

    def updated_iset(self, orm, val):
        if val:
            if not self.modified(orm, val):
                #val = datetime.datetime.utcnow()
                #val = datetime.datetime.now(datetime.timezone.utc)
                val = Datetime()

        else:
            #val = datetime.datetime.utcnow()
            #val = datetime.datetime.now(datetime.timezone.utc)
            val = Datetime()

        return val




# -*- coding: utf-8 -*-
from __future__ import unicode_literals, division, print_function, absolute_import

from datatypes import classproperty

from ..model import Orm
from ..config import (
    Field,
    Index,
)


class MagicOrm(Orm):
    """Extends the default Orm with some more syntactic sugar

    https://github.com/Jaymon/prom/issues/78
    """
    def __pout__(self):
        """This just makes the object easier to digest in pout.v() calls

        more information on what pout is: https://github.com/Jaymon/pout
        """
        return self.fields

    @classproperty
    def pk_name(cls):
        """return the preferred primary key name for the Orm

        you can access the primary key on an instance using ._id or .pk but usually
        when we have foreign keys we use modelname_id and then in the jsonable
        dicts that are passed down we include that modelname_id in the actual instance
        instead of an _id field (eg, the jsonable on object Foo would have 'foo_id' key)
        and this makes it easy to get that key name and override it if needed

        :returns: string, the preferred primary key name that is usually the name
            that should be used in jsonable and in foreign key fields on other
            models
        """
        return f"{cls.model_name}_id"

    def __getattr__(self, k):
        """Adds some syntactic sugar to the Orm

        adds support for:
            .<MODEL-NAME>_id = alias for self.pk or self._id
            .is_fieldname() = if fieldname is a boolean then returns True/False,
                if fieldname is another value then you can do .is_fieldname(val)
                to compare val to the fieldname's value
        """
        if k.startswith("is_"):
            field_name = k[3:]
            field = self.schema.fields[field_name]
            if issubclass(field.type, bool):
                ret = lambda: getattr(self, field_name)

            else:
                ret = lambda x: x == getattr(self, field_name)

        elif k == self.pk_name:
            ret = self.pk

        else:
            ret = super().__getattr__(k)

        return ret

    def jsonable(self):
        """Switches out _id for pk_name"""
        pk_fields = self.schema.pk_fields
        if len(pk_fields) == 1:
            for pk_field in pk_fields.values():
                pk_field.options.setdefault("jsonable_name", self.pk_name)

        return super().jsonable()

